use std::{marker::PhantomData, rc::Rc};

use crate::{
    flow::{binary::*, comparator::*},
    func::context::*,
    rcore::*,
    rstd::fallible::*,
};

use super::*;

struct TreeContext<C, T>(Rc<C>, PhantomData<T>);

impl<'a, Ctx: Context<'a>, A: Mentionable<'a, Ctx> + Clone, C: 'a + Comparator<A>, E: 'a>
    TreeContext<C, (Ctx, A, E)>
{
    fn new(comparator: Rc<C>) -> Self {
        Self(comparator, PhantomData)
    }
}

impl<'a, Ctx: Context<'a>, A: Mentionable<'a, Ctx> + Clone, C: 'a + Comparator<A>, E: 'a> Clone
    for TreeContext<C, (Ctx, A, E)>
{
    fn clone(&self) -> Self {
        Self::new(self.0.clone())
    }
}

impl<'a, Ctx: Context<'a>, A: Mentionable<'a, Ctx> + Clone, C: 'a + Comparator<A>, E: 'a>
    FunctorContext<'a> for TreeContext<C, (Ctx, A, E)>
{
    type T = FallibleMonad<'a, Ctx, Result<ResolutionFailure<'a, Ctx, Node<'a, Ctx, A>>, E>>;
}

impl<'a, Ctx: Context<'a>, A: Mentionable<'a, Ctx> + Clone, C: 'a + Comparator<A>, E: 'a>
    BinaryTrees<'a> for TreeContext<C, (Ctx, A, E)>
{
    type Node = Node<'a, Ctx, A>;

    type Reference = Point<'a, Ctx, Self::Node>;

    type Tree = Tree<'a, Ctx, A>;

    type Key = A;

    type Comparator = C;

    type _Tm = Self::T;

    fn comparator(&self) -> &Self::Comparator {
        &self.0
    }

    fn split(&self, node: &Self::Node) -> crate::flow::binary::Split<'a, Self> {
        (node.l.clone(), node.r.clone(), node.key.clone())
    }

    fn resolve(
        &self,
        reference: &Self::Reference,
    ) -> crate::flow::binary::BTWrap<'a, Self, Self::Node> {
        Ctx::stuff(Ctx::fmap(reference.resolve(), |resolved| {
            resolved.map(|rc| rc.as_ref().clone()).map_err(Ok)
        }))
    }

    fn equal(&self, rl: &Self::Reference, rr: &Self::Reference) -> bool {
        rl == rr
    }

    fn refer(&self, tree: &Self::Tree) -> Option<Self::Reference> {
        tree.node.as_ref().cloned()
    }
}
